#!/usr/bin/env python

"""
Copyright (c) 2016, Continuum Analytics, Inc.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of Continuum Analytics nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import print_function, division

import logging
import os
import pkg_resources
import re
import subprocess
import functools
from functools import lru_cache

from frozendict import frozendict, deepfreeze
import networkx as nx
from conda.models.match_spec import MatchSpec
from conda.models.records import PackageRecord
from conda_build import api
from conda_build.metadata import find_recipe, MetaData


log = logging.getLogger(__file__)
CONDA_BUILD_CACHE = os.environ.get("CONDA_BUILD_CACHE")
hash_length = api.Config().hash_length


# https://stackoverflow.com/questions/6358481/using-functools-lru-cache-with-dictionary-arguments
def freezeargs(func):
    """Convert a mutable dictionary into immutable.
    Useful to be compatible with cache
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        args = (frozendict(arg) if isinstance(arg, dict) else arg for arg in args)
        kwargs = {
            k: frozendict(v) if isinstance(v, dict) else v for k, v in kwargs.items()
        }
        return func(*args, **kwargs)

    return wrapped


def package_key(metadata, worker_label, run="build"):
    # get the build string from whatever conda-build makes of the configuration
    used_loop_vars = metadata.get_used_loop_vars()
    build_vars = "-".join(
        [
            k + "_" + str(metadata.config.variant[k])
            for k in used_loop_vars
            if k != "target_platform"
        ]
    )
    # kind of a special case.  Target platform determines a lot of output behavior, but may not be
    #    explicitly listed in the recipe.
    tp = metadata.config.variant.get("target_platform")
    if tp and tp != metadata.config.subdir and "target_platform" not in build_vars:
        build_vars += "-target_" + tp
    key = [metadata.name(), metadata.version()]
    if build_vars:
        key.append(build_vars)
    key.extend(["on", worker_label])
    key = "-".join(key)
    if run == "test":
        key = "-".join(("c3itest", key))
    return key


def _git_changed_files(git_rev, stop_rev=None, git_root=""):
    if not git_root:
        git_root = os.getcwd()
    if stop_rev:
        git_rev = "{0}..{1}".format(git_rev, stop_rev)
    print("Changed files from:", git_rev, stop_rev, git_root)
    output = subprocess.check_output(
        [
            "git",
            "-C",
            git_root,
            "diff-tree",
            "--no-commit-id",
            "--name-only",
            "-r",
            git_rev,
        ]
    )
    files = output.decode().splitlines()
    return files


def _get_base_folders(base_dir, changed_files):
    recipe_dirs = []
    for f in changed_files:
        # only consider files that come from folders
        if "/" in f:
            f = f.split("/")[0]
        try:
            find_recipe(os.path.join(base_dir, f))
            recipe_dirs.append(f)
        except IOError:
            pass
    return recipe_dirs


def git_changed_submodules(git_rev="HEAD@{1}", stop_rev=None, git_root="."):
    if stop_rev is not None:
        git_rev = "{0}..{1}".format(git_rev, stop_rev)
    diff_script = pkg_resources.resource_filename(
        "conda_concourse_ci", "diff-script.sh"
    )

    diff = subprocess.check_output(
        ["bash", diff_script, git_rev], cwd=git_root, universal_newlines=True
    )

    submodule_changed_files = [line.split() for line in diff.splitlines()]

    submodules_with_recipe_changes = []
    for submodule in submodule_changed_files:
        for file in submodule:
            if "recipe/" in file and submodule[0] not in submodules_with_recipe_changes:
                submodules_with_recipe_changes.append(submodule[0])

    return submodules_with_recipe_changes


def git_new_submodules(git_rev="HEAD@{1}", stop_rev=None, git_root="."):
    if stop_rev is not None:
        git_rev = "{0}..{1}".format(git_rev, stop_rev)

    new_submodule_script = pkg_resources.resource_filename(
        "conda_concourse_ci", "new-submodule-script.sh"
    )

    diff = subprocess.check_output(
        ["bash", new_submodule_script, git_rev], cwd=git_root, universal_newlines=True
    )

    return diff.splitlines()


def git_renamed_folders(git_rev="HEAD@{1}", stop_rev=None, git_root="."):
    if stop_rev is not None:
        git_rev = "{0}..{1}".format(git_rev, stop_rev)

    rename_script = pkg_resources.resource_filename(
        "conda_concourse_ci", "rename-script.sh"
    )

    renamed_files = subprocess.check_output(
        ["bash", rename_script], cwd=git_root, universal_newlines=True
    ).splitlines()

    return renamed_files


def git_changed_recipes(git_rev="HEAD@{1}", stop_rev=None, git_root="."):
    """
    Get the list of files changed in a git revision and return a list of
    package directories that have been modified.

    git_rev: if stop_rev is not provided, this represents the changes
             introduced by the given git rev.  It is equivalent to
             git_rev=SOME_REV@{1} and stop_rev=SOME_REV

    stop_rev: when provided, this is the end of a range of revisions to
             consider.  git_rev becomes the start revision.  Note that the
             start revision is *one before* the actual start of examining
             commits for changes.  In other words:

             git_rev=SOME_REV@{1} and stop_rev=SOME_REV   => only SOME_REV
             git_rev=SOME_REV@{2} and stop_rev=SOME_REV   => two commits, SOME_REV and the
                                                             one before it
    """
    changed_files = _git_changed_files(git_rev, stop_rev, git_root)
    recipe_dirs = _get_base_folders(git_root, changed_files)
    changed_submodules = git_changed_submodules(git_rev, stop_rev, git_root)
    new_submodules = git_new_submodules(git_rev, stop_rev, git_root)
    renamed_folders = git_renamed_folders(git_rev, stop_rev, git_root)
    return recipe_dirs + changed_submodules + new_submodules + renamed_folders


def _deps_to_version_dict(deps):
    d = {}
    for x in deps:
        x = x.strip().split()
        if len(x) == 3:
            d[x[0]] = (x[1], x[2])
        elif len(x) == 2:
            d[x[0]] = (x[1], "any")
        else:
            d[x[0]] = ("any", "any")
    return d


def get_build_deps(meta):
    build_reqs = meta.get_value("requirements/build")
    if not build_reqs:
        build_reqs = []
    return _deps_to_version_dict(build_reqs)


def get_run_test_deps(meta):
    run_reqs = meta.get_value("requirements/run")
    if not run_reqs:
        run_reqs = []
    test_reqs = meta.get_value("test/requires")
    if not test_reqs:
        test_reqs = []
    return _deps_to_version_dict(run_reqs + test_reqs)


_rendered_recipes = {}


@freezeargs
@lru_cache(maxsize=None)
def _get_or_render_metadata(meta_file_or_recipe_dir, worker, finalize, config=None):
    global _rendered_recipes
    platform = worker["platform"]
    arch = str(worker["arch"])
    if (meta_file_or_recipe_dir, platform, arch) not in _rendered_recipes:
        print("rendering {0} for {1}".format(meta_file_or_recipe_dir, worker["label"]))
        _rendered_recipes[(meta_file_or_recipe_dir, platform, arch)] = api.render(
            meta_file_or_recipe_dir,
            platform=platform,
            arch=arch,
            verbose=False,
            permit_undefined_jinja=True,
            bypass_env_check=True,
            config=config,
            finalize=finalize,
        )
    return _rendered_recipes[(meta_file_or_recipe_dir, platform, arch)]


def add_recipe_to_graph(
    recipe_dir,
    graph,
    run,
    worker,
    conda_resolve,
    recipes_dir=None,
    config=None,
    finalize=False,
):
    try:
        print(recipe_dir, worker, config, finalize, flush=True)
        rendered = _get_or_render_metadata(
            recipe_dir, worker, config=config, finalize=finalize
        )
    except (IOError, SystemExit):
        log.exception("invalid recipe dir: %s", recipe_dir)
        raise

    name = None
    for metadata, _, _ in rendered:
        name = package_key(metadata, worker["label"], run)

        if metadata.skip():
            continue

        if name not in graph.nodes():
            graph.add_node(name, meta=metadata, worker=worker)
            add_dependency_nodes_and_edges(
                name,
                graph,
                run,
                worker,
                conda_resolve,
                config=config,
                recipes_dir=recipes_dir,
                finalize=finalize,
            )

        # # add the test equivalent at the same time.  This is so that expanding can find it.
        # if run == 'build':
        #     add_recipe_to_graph(recipe_dir, graph, 'test', worker, conda_resolve,
        #                         recipes_dir=recipes_dir)
        #     test_key = package_key(metadata, worker['label'])
        #     graph.add_edge(test_key, name)
        #     upload_key = package_key(metadata, worker['label'])
        #     graph.add_node(upload_key, meta=metadata, worker=worker)
        #     graph.add_edge(upload_key, test_key)

    return name


def match_peer_job(target_matchspec, other_m, this_m=None):
    """target_matchspec comes from the recipe.  target_variant is the variant from the recipe whose
    deps we are matching.  m is the peer job, which must satisfy conda and also have matching keys
    for any keys that are shared between target_variant and m.config.variant"""
    match_dict = {
        "name": other_m.name(),
        "version": other_m.version(),
        "build": _fix_any(other_m.build_id(), other_m.config),
    }
    match_record = PackageRecord(
        name=match_dict["name"],
        version=match_dict["version"],
        build=match_dict["build"],
        build_number=int(other_m.build_number() or 0),
        channel=None,
    )
    matchspec_matches = target_matchspec.match(match_record)

    variant_matches = True
    if this_m:
        other_m_used_vars = other_m.get_used_loop_vars()
        for v in this_m.get_used_loop_vars():
            if v in other_m_used_vars:
                variant_matches &= this_m.config.variant[v] == other_m.config.variant[v]
    return matchspec_matches and variant_matches


def add_intradependencies(graph):
    """ensure that downstream packages wait for upstream build/test (not use existing
    available packages)"""
    for node in graph.nodes():
        if "meta" not in graph.nodes[node]:
            continue
        # get build dependencies
        m = graph.nodes[node]["meta"]
        # this is pretty hard. Realistically, we would want to know
        # what the build and host platforms are on the build machine.
        # However, all we know right now is what machine we're actually
        # on (the one calculating the graph).

        test_requires = m.meta.get("test", {}).get("requires", [])

        log.info("node: {}".format(node))
        log.info("   build: {}".format(m.ms_depends("build")))
        log.info("   host: {}".format(m.ms_depends("host")))
        log.info("   run: {}".format(m.ms_depends("run")))
        log.info("   test: {}".format(test_requires))

        deps = set(
            m.ms_depends("build")
            + m.ms_depends("host")
            + m.ms_depends("run")
            + [MatchSpec(dep) for dep in test_requires or []]
        )

        for dep in deps:
            name_matches = (
                n for n in graph.nodes() if graph.nodes[n]["meta"].name() == dep.name
            )
            for matching_node in name_matches:
                # are any of these build dependencies also nodes in our graph?
                if (
                    match_peer_job(
                        MatchSpec(dep), graph.nodes[matching_node]["meta"], m
                    )
                    and (node, matching_node) not in graph.edges()
                ):
                    # add edges if they don't already exist
                    graph.add_edge(node, matching_node)


def collapse_subpackage_nodes(graph):
    """Collapse all subpackage nodes into their parent recipe node

    We get one node per output, but a given recipe can have multiple outputs.  It's important
    for dependency ordering in the graph that the outputs exist independently, but once those
    dependencies are established, we need to collapse subpackages down to a single job for the
    top-level recipe."""
    # group nodes by their recipe path first, then within those groups by their variant
    node_groups = {}
    for node in graph.nodes():
        if "meta" in graph.nodes[node]:
            meta = graph.nodes[node]["meta"]
            meta_path = meta.meta_path or meta.meta["extra"]["parent_recipe"]["path"]
            master = False

            master_meta = MetaData(meta_path, config=meta.config)
            if master_meta.name() == meta.name():
                master = True
            group = node_groups.get(meta_path, {})
            subgroup = group.get(deepfreeze(meta.config.variant), {})
            if master:
                if "master" in subgroup:
                    raise ValueError(
                        "tried to set more than one node in a group as master"
                    )
                subgroup["master"] = node
            else:
                sps = subgroup.get("subpackages", [])
                sps.append(node)
                subgroup["subpackages"] = sps
            group[deepfreeze(meta.config.variant)] = subgroup
            node_groups[meta_path] = group

    for recipe_path, group in node_groups.items():
        for variant, subgroup in group.items():
            # if no node is the top-level recipe (only outputs, no top-level output), need to obtain
            #     package/name from recipe given by common recipe path.
            subpackages = subgroup.get("subpackages")
            if "master" not in subgroup:
                sp0 = graph.nodes[subpackages[0]]
                master_meta = MetaData(recipe_path, config=sp0["meta"].config)
                worker = sp0["worker"]
                master_key = package_key(master_meta, worker["label"])
                graph.add_node(master_key, meta=master_meta, worker=worker)
                master = graph.nodes[master_key]
            else:
                master = subgroup["master"]
                master_key = package_key(
                    graph.nodes[master]["meta"], graph.nodes[master]["worker"]["label"]
                )
            # fold in dependencies for all of the other subpackages within a group.  This is just
            #     the intersection of the edges between all nodes.  Store this on the "master" node.
            if subpackages:
                remap_edges = [edge for edge in graph.edges() if edge[1] in subpackages]
                for edge in remap_edges:
                    # make sure not to add references to yourself
                    if edge[0] != master_key:
                        graph.add_edge(edge[0], master_key)
                    graph.remove_edge(*edge)

                # remove nodes that have been folded into master nodes
                for subnode in subpackages:
                    graph.remove_node(subnode)


def construct_graph(
    recipes_dir,
    worker,
    run,
    conda_resolve,
    folders=(),
    git_rev=None,
    stop_rev=None,
    matrix_base_dir=None,
    config=None,
    finalize=False,
):
    """
    Construct a directed graph of dependencies from a directory of recipes

    run: whether to use build or run/test requirements for the graph.  Avoids cycles.
          values: 'build' or 'test'.  Actually, only 'build' matters - otherwise, it's
                   run/test for any other value.
    """
    matrix_base_dir = matrix_base_dir or recipes_dir
    if not os.path.isabs(recipes_dir):
        recipes_dir = os.path.normpath(os.path.join(os.getcwd(), recipes_dir))
    assert os.path.isdir(recipes_dir)

    if not folders:
        if not git_rev:
            git_rev = "HEAD"

        folders = git_changed_recipes(git_rev, stop_rev=stop_rev, git_root=recipes_dir)

    graph = nx.DiGraph()
    for folder in folders:
        recipe_dir = os.path.join(recipes_dir, folder)
        if not os.path.isdir(recipe_dir):
            raise ValueError("Specified folder {} does not exist".format(recipe_dir))
        add_recipe_to_graph(
            recipe_dir,
            graph,
            run,
            worker,
            conda_resolve,
            recipes_dir,
            config=config,
            finalize=finalize,
        )
    add_intradependencies(graph)
    collapse_subpackage_nodes(graph)
    return graph


def _fix_any(value, config):
    value = re.sub("any(?:h[0-9a-f]{%d})?" % config.hash_length, "", value)
    return value


@lru_cache(maxsize=None)
def _installable(name, version, build_string, config, conda_resolve):
    """Can Conda install the package we need?"""
    ms = MatchSpec(
        " ".join([name, _fix_any(version, config), _fix_any(build_string, config)])
    )
    installable = conda_resolve.find_matches(ms)
    if not installable:
        log.warn(
            "Dependency {name}, version {ver} is not installable from your "
            "channels: {channels} with subdir {subdir}.  Seeing if we can build it...".format(
                name=name,
                ver=version,
                channels=config.channel_urls,
                subdir=config.host_subdir,
            )
        )
    return installable


def _buildable(name, version, recipes_dir, worker, config, finalize):
    """Does the recipe that we have available produce the package we need?"""
    possible_dirs = os.listdir(recipes_dir)
    packagename_re = re.compile(r"%s(?:\-[0-9]+[\.0-9\_\-a-zA-Z]*)?$" % name)
    likely_dirs = (
        dirname
        for dirname in possible_dirs
        if (
            os.path.isdir(os.path.join(recipes_dir, dirname))
            and packagename_re.match(dirname)
        )
    )
    metadata_tuples = [
        m
        for path in likely_dirs
        for (m, _, _) in _get_or_render_metadata(
            os.path.join(recipes_dir, path), worker, finalize=finalize
        )
    ]

    # this is our target match
    ms = MatchSpec(" ".join([name, _fix_any(version, config)]))
    available = False
    for m in metadata_tuples:
        available = match_peer_job(ms, m)
        if available:
            break
    return m.meta_path if available else False


def add_dependency_nodes_and_edges(
    node,
    graph,
    run,
    worker,
    conda_resolve,
    recipes_dir=None,
    finalize=False,
    config=None,
):
    """add build nodes for any upstream deps that are not yet installable

    changes graph in place.
    """
    metadata = graph.nodes[node]["meta"]
    # for plain test runs, ignore build reqs.
    deps = get_run_test_deps(metadata)
    recipes_dir = recipes_dir or os.getcwd()

    # cross: need to distinguish between build_subdir (build reqs) and host_subdir
    if run == "build":
        deps.update(get_build_deps(metadata))

    for dep, (version, build_str) in deps.items():
        # we don't need worker info in _installable because it is already part of conda_resolve
        if not _installable(dep, version, build_str, metadata.config, conda_resolve):
            recipe_dir = _buildable(
                dep, version, recipes_dir, worker, metadata.config, finalize=finalize
            )
            if not recipe_dir:
                continue
                # raise ValueError("Dependency {} is not installable, and recipe (if "
                #                  " available) can't produce desired version ({})."
                #                  .format(dep, version))
            dep_name = add_recipe_to_graph(
                recipe_dir,
                graph,
                "build",
                worker,
                conda_resolve,
                recipes_dir,
                config=config,
                finalize=finalize,
            )
            if not dep_name:
                raise ValueError(
                    "Tried to build recipe {0} as dependency, which is skipped "
                    "in meta.yaml".format(recipe_dir)
                )
            graph.add_edge(node, dep_name)


def expand_run_upstream(
    graph,
    conda_resolve,
    worker,
    run,
    steps=0,
    max_downstream=5,
    recipes_dir=None,
    matrix_base_dir=None,
):
    pass


def expand_run(
    graph,
    conda_resolve,
    worker,
    run,
    steps=0,
    max_downstream=5,
    recipes_dir=None,
    matrix_base_dir=None,
    finalize=False,
):
    """Apply the build label to any nodes that need (re)building or testing.

    "need rebuilding" means both packages that our target package depends on,
    but are not yet built, as well as packages that depend on our target
    package. For the latter, you can specify how many dependencies deep (steps)
    to follow that chain, since it can be quite large.

    If steps is -1, all downstream dependencies are rebuilt or retested
    """
    downstream = 0
    initial_nodes = len(graph.nodes())

    # for build, we get test automatically.  Give people the max_downstream in terms
    #   of packages, not tasks
    # if run == 'build':
    #     max_downstream *= 2

    def expand_step(task_graph, full_graph, downstream):
        for node in task_graph.nodes():
            for predecessor in full_graph.predecessors(node):
                if max_downstream < 0 or (downstream - initial_nodes) < max_downstream:
                    add_recipe_to_graph(
                        os.path.dirname(
                            full_graph.nodes[predecessor]["meta"].meta_path
                        ),
                        task_graph,
                        run=run,
                        worker=worker,
                        conda_resolve=conda_resolve,
                        recipes_dir=recipes_dir,
                        finalize=finalize,
                    )
                    downstream += 1
        return len(graph.nodes())

    # starting from our initial collection of dirty nodes, trace the tree down to packages
    #   that depend on the dirty nodes.  These packages may need to be rebuilt, or perhaps
    #   just tested.  The 'run' argument determines which.

    if steps != 0:
        if not recipes_dir:
            raise ValueError(
                "recipes_dir is necessary if steps != 0.  "
                "Please pass it as an argument."
            )
        # here we need to fully populate a graph that has the right build or run/test deps.
        #    We don't create this elsewhere because it is unnecessary and costly.

        # get all immediate subdirectories
        other_top_dirs = [
            d
            for d in os.listdir(recipes_dir)
            if os.path.isdir(os.path.join(recipes_dir, d)) and not d.startswith(".")
        ]
        recipe_dirs = []
        for recipe_dir in other_top_dirs:
            try:
                find_recipe(os.path.join(recipes_dir, recipe_dir))
                recipe_dirs.append(recipe_dir)
            except IOError:
                pass

        # constructing the graph for build will automatically also include the test deps
        full_graph = construct_graph(
            recipes_dir,
            worker,
            "build",
            folders=recipe_dirs,
            matrix_base_dir=matrix_base_dir,
            conda_resolve=conda_resolve,
        )

        if steps >= 0:
            for step in range(steps):
                downstream = expand_step(graph, full_graph, downstream)
        else:
            while True:
                nodes = graph.nodes()
                downstream = expand_step(graph, full_graph, downstream)
                if nodes == graph.nodes():
                    break


def order_build(graph):
    """
    Assumes that packages are in graph.
    Builds a temporary graph of relevant nodes and returns it topological sort.

    Relevant nodes selected in a breadth first traversal sourced at each pkg
    in packages.
    """
    reorder_cyclical_test_dependencies(graph)
    try:
        order = list(nx.topological_sort(graph))
        order.reverse()
    except nx.exception.NetworkXUnfeasible:
        raise ValueError(
            "Cycles detected in graph: %s", nx.find_cycle(graph, orientation="reverse")
        )

    return order


def reorder_cyclical_test_dependencies(graph):
    """By default, we make things that depend on earlier outputs for build wait for tests of
    the earlier thing to pass.  However, circular dependencies spread across run/test and
    build/host can make this approach incorrect. For example:

    A <-- B  : B depends on A at build time
    B <-- A  : A depends on B at run time.  We can build A before B, but we cannot test A until B
               is built.

    To resolve this, we must reorder the graph edges:

    build A <-- test A <--> build B  <-- test B

    must become:

    build A  <-- build B <-- test A <-- test B
    """
    # find all test nodes with edges to build nodes
    test_nodes = [node for node in graph.nodes() if node.startswith("test-")]
    edges_from_test_to_build = [
        edge
        for edge in graph.edges()
        if edge[0] in test_nodes and edge[1].startswith("build-")
    ]

    # find any of their inverses.  Entries here are of the form (test-A, build-B)
    circular_deps = [
        edge for edge in edges_from_test_to_build if (edge[1], edge[0]) in graph.edges()
    ]

    for testA, buildB in circular_deps:
        # remove build B dependence on test A
        graph.remove_edge(testA, buildB)
        # remove test B dependence on build B
        testB = buildB.replace("build-", "test-", 1)
        graph.remove_edge(buildB, testB)
        # Add test B dependence on test A
        graph.add_edge(testA, testB)
        # make sure that test A still depends on build B
        assert (buildB, testA) in graph.edges()
    # graph is modified in place.  No return necessary.



import argparse
import os

import github
import requests


def _get_latest_run_summary(repo, workflow_run_id):
    latest_run = repo.get_workflow_run(workflow_run_id)
    for job in latest_run.jobs():
        pass

    r = requests.get(job.logs_url())

    summary = ""
    in_summary = False
    for line in r.text.splitlines():
        line = line.strip()
        if "###START-OF-SUMMARY###" in line:
            in_summary = True
            head_len = len(line.split("###START-OF-SUMMARY###")[0])
            continue

        if "###END-OF-SUMMARY###" in line:
            in_summary = False
            break

        if in_summary:
            line = line[head_len:]
            summary += line + "\n"

    return summary


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Lint staged recipes.")
    parser.add_argument(
        "--head-repo-owner", type=str, required=True, help="the head repo owner"
    )
    parser.add_argument(
        "--workflow-run-id", type=int, required=True, help="the ID of the workflor run"
    )
    parser.add_argument(
        "--head-sha", type=str, required=True, help="the head SHA of the PR"
    )

    args = parser.parse_args()

    gh = github.Github(auth=github.Auth.Token(os.environ["GH_TOKEN"]))
    head_repo = gh.get_repo(f"{args.head_repo_owner}/staged-recipes")
    base_repo = gh.get_repo("conda-forge/staged-recipes")

    summary = _get_latest_run_summary(base_repo, args.workflow_run_id)
    if summary:
        print(summary)
        commit = head_repo.get_commit(args.head_sha)
        pr = None
        for _pr in commit.get_pulls():
            if _pr.base.repo.full_name == base_repo.full_name:
                pr = _pr
                break

        if pr is None:
            # for reasons I do not follow, sometimes the head commit API
            # to get pull requests does not return the PR.
            # So we try looking at PRs from base repo and find the same
            # sha+head repo but limit the search since staged-recipes
            # gets tons of PRs.
            max_tries = 50
            num_tries = 0
            for _pr in base_repo.get_pulls():
                if (
                    _pr.head.sha == args.head_sha
                    and _pr.head.repo.full_name == head_repo.full_name
                ):
                    pr = _pr
                    break
                num_tries += 1
                if num_tries == max_tries:
                    break

        if pr is not None:
            comment = None
            for _comment in pr.get_issue_comments():
                if "Hi! This is the staged-recipes linter" in _comment.body:
                    comment = _comment

            if comment:
                if comment.body != summary:
                    curr_fline = comment.body.splitlines()[0].strip()
                    new_fl = summary.splitlines()[0].strip()
                    if curr_fline == new_fl:
                        comment.edit(summary)
                    else:
                        pr.create_issue_comment(summary)
            else:
                pr.create_issue_comment(summary)
        else:
            print("No PR found for the given commit. No comment being made!")



import argparse
import os
import sys
from collections import defaultdict
from pathlib import Path
from subprocess import check_output

from conda_smithy.linter import conda_recipe_v1_linter
from conda_smithy.linter.utils import (
    get_section,
)
from conda_smithy.utils import get_yaml, render_meta_yaml
import github
import requests

NOCOMMENT_REQ_TEAMS = ["conda-forge/r", "conda-forge/cuda"]
HERE = Path(__file__).parent
ROOT = (HERE / ".." / ".." / "..").absolute()


def _test_and_raise_besides_file_not_exists(e: github.GithubException):
    if isinstance(e, github.UnknownObjectException):
        return
    if e.status == 404 and "No object found" in e.data["message"]:
        return
    raise e


def _lint_recipes(gh, pr):
    lints = defaultdict(list)
    hints = defaultdict(list)
    fnames = set(f.filename for f in pr.get_files())
    labels = set(label.name for label in pr.get_labels())
    extra_edits = False
    example_recipes = ["recipes/example/meta.yaml", "recipes/example-v1/recipe.yaml"]

    # 1. Do not edit or delete example recipes and only edit recipe files
    if "maintenance" not in labels:
        for fname in fnames:
            if fname in example_recipes:
                lints[fname].append(
                    "Do not edit or delete example recipes in `recipes/example/` or `recipe/example-v1/`."
                )
                extra_edits = True
            if not fname.startswith("recipes/"):
                lints[fname].append(
                    "Do not edit files outside of the `recipes/` directory."
                )
                extra_edits = True

    # 2. Make sure the new recipe is in the right directory
    for fname in fnames:
        if (
            fname.startswith("recipes/example/")
            or fname.startswith("recipes/example-v1/")
            or fname.startswith("recipes/meta.y")
            or fname.startswith("recipes/recipe.y")
        ) and fname not in example_recipes:
            lints[fname].append(
                "Please put your recipe in its own directory in the `recipes/` directory as "
                "`recipe/<name of feedstock>/<your recipe file>.yaml`."
            )

    # 3. Ensure environment.yaml and pixi.toml are in sync
    original_environment_yaml = (ROOT / "environment.yaml").read_text().rstrip()
    pixi_exported_env_yaml = check_output(
        ["pixi", "project", "export", "conda-environment", "-e", "build"],
        text=True,
    ).rstrip()
    if original_environment_yaml != pixi_exported_env_yaml:
        import difflib

        _orig_lines = original_environment_yaml.splitlines(keepends=True)
        _expt_lines = pixi_exported_env_yaml.splitlines(keepends=True)
        print("environment diff:", flush=True)
        print("".join(difflib.unified_diff(_orig_lines, _expt_lines)), flush=True)
        lints["environment.yaml"].append(
            "The `environment.yaml` file is out of sync with `pixi.toml`. "
            "Fix by running `pixi project export conda-environment -e build > environment.yaml`."
        )

    # Recipe-specific lints/hints
    for fname in fnames:
        if (
            (not fname.startswith("recipes/"))
            or (not fname.endswith(".yaml"))
            or fname in example_recipes
        ):
            continue

        # grab basic metadata
        if fname.endswith("meta.yaml"):
            with open(fname) as fh:
                content = render_meta_yaml("".join(fh))
                meta = get_yaml().load(content)
            recipe_version = 0
        else:
            meta = get_yaml().load(Path(fname))
            recipe_version = 1

        package_section = get_section(
            meta, "package", lints, recipe_version=recipe_version
        )
        sources_section = get_section(
            meta, "source", lints, recipe_version=recipe_version
        )
        outputs_section = get_section(
            meta, "outputs", lints, recipe_version=recipe_version
        )
        extra_section = get_section(meta, "extra", lints, recipe_version=recipe_version)
        maintainers = extra_section.get("recipe-maintainers", [])

        if recipe_version == 1:
            recipe_name = conda_recipe_v1_linter.get_recipe_name(meta)
        else:
            recipe_name = package_section.get("name", "").strip()

        recipe_name = extra_section.get("feedstock-name", recipe_name)

        # 4. Check for existing feedstocks in conda-forge or bioconda
        if recipe_name:
            cf = gh.get_user("conda-forge")

            for name in set(
                [
                    recipe_name,
                    recipe_name.replace("-", "_"),
                    recipe_name.replace("_", "-"),
                ]
            ):
                try:
                    if cf.get_repo(f"{name}-feedstock"):
                        existing_recipe_name = name
                        feedstock_exists = True
                        break
                    else:
                        feedstock_exists = False
                except github.GithubException as e:
                    _test_and_raise_besides_file_not_exists(e)
                    feedstock_exists = False

            if feedstock_exists and existing_recipe_name == recipe_name:
                lints[fname].append(
                    "Feedstock with the same name exists in conda-forge."
                )
            elif feedstock_exists:
                lints[fname].append(
                    f"Feedstock with the name {existing_recipe_name} exists in conda-forge. "
                    f"Is it the same as this package ({recipe_name})?"
                )

            bio = gh.get_user("bioconda").get_repo("bioconda-recipes")
            try:
                bio.get_dir_contents(f"recipes/{recipe_name}")
            except github.GithubException as e:
                _test_and_raise_besides_file_not_exists(e)
            else:
                lints[fname].append(
                    "Recipe with the same name exists in bioconda: "
                    "please discuss with @conda-forge/bioconda-recipes."
                )

            url = None
            if recipe_version == 1:
                for source_section in sources_section:
                    if str(source_section.get("url")).startswith(
                        "https://pypi.io/packages/source/"
                    ):
                        url = source_section["url"]
            else:
                for source_section in sources_section:
                    if str(source_section.get("url")).startswith(
                        "https://pypi.io/packages/source/"
                    ):
                        url = source_section["url"]
            if url:
                # get pypi name from  urls like "https://pypi.io/packages/source/b/build/build-0.4.0.tar.gz"
                pypi_name = url.split("/")[6]
                mapping_request = requests.get(
                    "https://raw.githubusercontent.com/regro/cf-graph-countyfair/master/mappings/pypi/name_mapping.yaml"
                )
                if mapping_request.status_code == 200:
                    mapping_raw_yaml = mapping_request.content
                    mapping = get_yaml().load(mapping_raw_yaml)
                    for pkg in mapping:
                        if pkg.get("pypi_name", "") == pypi_name:
                            conda_name = pkg["conda_name"]
                            lints[fname].append(
                                f"A conda package with same name ({conda_name}) already exists."
                            )

        # 5. Ensure all maintainers have commented that they approve of being listed
        if maintainers:
            # Get PR author, issue comments, and review comments
            pr_author = pr.user.login.lower()
            issue_comments = pr.get_issue_comments()
            review_comments = pr.get_reviews()

            # Combine commenters from both issue comments and review comments
            commenters = {comment.user.login.lower() for comment in issue_comments}
            commenters.update({review.user.login.lower() for review in review_comments})

            # Check if all maintainers have either commented or are the PR author
            non_participating_maintainers = set()
            for orig_maintainer in maintainers:
                maintainer = orig_maintainer.lower()
                if (
                    maintainer not in commenters
                    and maintainer != pr_author
                    and maintainer not in NOCOMMENT_REQ_TEAMS
                    and "/" not in maintainer
                ):
                    non_participating_maintainers.add(orig_maintainer)

            # Add a lint message if there are any non-participating maintainers
            if non_participating_maintainers:
                lints[fname].append(
                    f"The following maintainers have not yet confirmed that they are willing to be listed here: "
                    f"{', '.join(non_participating_maintainers)}. Please ask them to comment on this PR if they are."
                )

        # 6. Only conda-forge teams can be maintainers
        if maintainers:
            for maintainer in maintainers:
                if "/" in maintainer:
                    res = maintainer.split("/", 1)
                    if len(res) == 2:
                        org, team = res
                        if org != "conda-forge":
                            lints[fname].append(
                                "Only conda-forge teams can be team maintainers."
                                f" {maintainer} is not a team in conda-forge."
                            )
                    else:
                        lints[fname].append(
                            f'Maintainer team "{maintainer}" is not in the '
                            'correct format. Please use "org/team" format.'
                        )

        # 7. Check that feedstock-name is defined if recipe is multi-output
        if outputs_section and ("feedstock-name" not in extra_section):
            hints[fname].append(
                "It looks like you are submitting a multi-output recipe. "
                "In these cases, the correct name for the feedstock is ambiguous, "
                "and our infrastructure defaults to the top-level `package.name` field. "
                "Please add a `feedstock-name` entry in the `extra` section."
            )

    return dict(lints), dict(hints), extra_edits


def _comment_on_pr(pr, lints, hints, extra_edits):
    if lints:
        topline = "I found some lint."
    elif hints:
        topline = "your PR looks excellent but I have some suggestions."
    else:
        topline = "your PR looks excellent! :rocket:"
    summary = f"Hi! This is the staged-recipes linter and {topline}\n"

    if extra_edits:
        summary += """\
\nIt looks like some changes were made outside the `recipes/` directory. \
To ensure everything runs smoothly, please make sure that recipes are only \
added to the `recipes/` directory and no other files are changed.

If these changes are intentional (and you aren't submitting a recipe), \
please add a `maintenance` label to the PR.\n"""

    added_lint_hint_header = False
    all_fnames = set(lints.keys()) | set(hints.keys())
    for fname in all_fnames:
        lint_message = ""
        hint_message = ""

        if fname in lints and lints[fname]:
            lint_message = "  - lints:\n"
            for lint in lints[fname]:
                if lint:
                    lint_message += f"    - {lint}\n"

        if fname in hints and hints[fname]:
            hint_message = "  - hints:\n"
            for hint in hints[fname]:
                if hint:
                    hint_message += f"    - {hint}\n"

        if lint_message or hint_message:
            if not added_lint_hint_header:
                summary += "\nFile-specific lints and/or hints:\n"
                added_lint_hint_header = True
            summary += f"\n- `{fname}`:\n"
            summary += lint_message + hint_message + "\n"

    print(summary)
    print("###START-OF-SUMMARY###", file=sys.stderr)
    print(summary, file=sys.stderr)
    print("###END-OF-SUMMARY###", file=sys.stderr)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Lint staged recipes.")
    parser.add_argument("--owner", type=str, required=True, help="the repo owner")
    parser.add_argument("--pr-num", type=int, required=True, help="the PR number")

    args = parser.parse_args()

    gh = github.Github(auth=github.Auth.Token(os.getenv("GH_TOKEN")))
    repo = gh.get_repo(f"{args.owner}/staged-recipes")
    pr = repo.get_pull(args.pr_num)

    lints, hints, extra_edits = _lint_recipes(gh, pr)
    _comment_on_pr(pr, lints, hints, extra_edits)
    if lints:
        sys.exit(1)



import numpy as np
from pyhough.pm import *

def hfdf_hough_spectrogram(times,freqs,power,Tsft,sdgrid,threshold,vec_n,vels,ref_perc_time):
#### times: unique times array
#### freqs: unique freqs array
#### power: spectrogram of size len(freqs) x len(times)
    inif = np.min(freqs)
    finf = np.max(freqs)
    df = 1/Tsft
    Nsds = len(sdgrid)
    Ntts = len(times)

    Nf0s = int(np.ceil((finf-inif)/df))

    binh_df0 = np.zeros((Nsds,Nf0s))
    t0 = np.percentile(times,ref_perc_time)

    times = times - t0
    for t in range(Ntts):
        inds_above_thr = np.array(select_local_max(power[:,t],threshold))
        if inds_above_thr.shape[0] == 0:
            continue
        
        freqs_peaks = freqs[inds_above_thr]
        freqs_dopp_corr = remove_doppler(freqs_peaks,vec_n,vels[:,t])
        these_fs = ( freqs_dopp_corr - inif ) / df ;
        tt_df = times[t] / df
        for k in range(Nsds):
            td = sdgrid[k] * tt_df
            inds = np.round(these_fs - td).astype(int) ### f = f0 + fdot(t-t0) solve for f0, f0 = f - fdot(t-t0)
            ind_of_inds = np.argwhere(inds>=0);
            a = inds[ind_of_inds];
            log_inds = a <= Nf0s-1;
            a = a[log_inds];
            binh_df0[k,a] = binh_df0[k,a] + 1
    
    return binh_df0


def hfdf_hough(times,peak_freqs,Tsft,sdgrid,ref_perc_time):

    inif = np.min(peak_freqs)
    finf = np.max(peak_freqs)
    df = 1/Tsft
    Nsds = len(sdgrid)
    Nf0s = int(np.ceil((finf-inif)/df))

    binh_df0 = np.zeros((Nsds,Nf0s))
    t0 = np.percentile(times,ref_perc_time)

    ii0 = 0
    ii = np.squeeze(np.argwhere(np.diff(times)))
    times = times - t0
    Ntts = len(ii)
    for t in range(Ntts):
        these_fs = ( peak_freqs[ii0:ii[t]+1] - inif ) / df 
        tt_df = times[ii0] / df
        for k in range(Nsds):
            td = sdgrid[k] * tt_df
            inds = np.round(these_fs - td).astype(int) ### f = f0 + fdot(t-t0) solve for f0, f0 = f - fdot(t-t0)
            ind_of_inds = np.argwhere(inds>=0);
            a = inds[ind_of_inds];
            log_inds = a <= Nf0s-1;
            a = a[log_inds];
            binh_df0[k,a] = binh_df0[k,a] + 1
            
            
        ii0 = ii[t]+1
    return binh_df0

def make_sd_grid(sig_fdot,dsd):
    min_fdot_search = sig_fdot * 10
    max_fdot_search = np.abs(sig_fdot * 10)
    if min_fdot_search == 0:
        min_fdot_search = -10 * dsd
        max_fdot_search = 10 * dsd
    sdgrid = np.arange(min_fdot_search,max_fdot_search,dsd)
    return sdgrid



import numpy as np
import pyfstat
import matplotlib.pyplot as plt

def get_detector_velocities(data):
    states = pyfstat.DetectorStates().get_multi_detector_states_from_sfts(data.sftfilepath, central_frequency=data.F0, time_offset=0)
    ts = np.array([data.tGPS.gpsSeconds for data in states.data[0].data])
    velocities = np.vstack([data.vDetector for data in states.data[0].data]).T
    return ts, velocities
    
    
def astro2rect(a, icrad=0):
    # Conversion from astronomical to rectangular coordinates
    # position = [signal_parameters["Alpha"], signal_parameters["Delta"]]
    # r = astro2rect(position,1) #icrad = 1: inputs are in rads; 0: inputs degs

    if len(a) == 2:
        a = np.append(a, 1)

    if icrad == 0:
        deg2rad = np.pi / 180
        a[0] = a[0] * deg2rad
        a[1] = a[1] * deg2rad

    r = np.zeros(3)
    r[0] = np.cos(a[0]) * np.cos(a[1]) * a[2]
    r[1] = np.sin(a[0]) * np.cos(a[1]) * a[2]
    r[2] = np.sin(a[1]) * a[2]

    return r
    
    
def remove_doppler(freqs,vec_n,velocities):

    freqs_dopp_corr = freqs / (1 + np.dot(vec_n, velocities))
    
    return freqs_dopp_corr

def python_plot_triplets(x, y, z, marker, label='', flag_logx=0, flag_logy=0, flag_log_cb=0, colorm='inferno', size=10, show_colorbar=True):
    """
    Plot triplets (x, y, z) with colormap encoding z values.

    Parameters:
    - x, y, z: Data arrays
    - marker: Marker style for plotting
    - label: Label for the colorbar
    - flag_logx, flag_logy: Flags to apply log scale to x or y axes
    - flag_log_cb: Flag to apply log scaling to colorbar values
    - colorm: Colormap to use
    - size: Size of the markers
    - show_colorbar: Whether to display the colorbar (True/False)
    """
    if flag_log_cb == 1:
        z = np.log10(z)

    # Define colormap and normalization
    cmap = plt.get_cmap(colorm)
    norm = plt.Normalize(vmin=np.min(z), vmax=np.max(z))

    # Create the figure and axis
    fig, ax = plt.subplots()

    # Scatter plot with color mapping
    sc = ax.scatter(x, y, c=z, cmap=cmap, norm=norm, marker=marker, s=size)

    # Handle log scales
    if flag_logx:
        ax.set_xscale('log')
    if flag_logy:
        ax.set_yscale('log')

    # Add colorbar if requested
    if show_colorbar:
        cbar = fig.colorbar(sc, ax=ax)
        cbar.set_label(label)

    return fig, ax


def make_peakmap_from_spectrogram(alltimes,freqs,normalized_power,threshold=3):
    Ntts = len(alltimes)
    freqs_new = []
    times_new = []
    powss_new = []
    index = [0]
    num_peaks = 0
    for t in range(Ntts):
        inds_above_thr = np.array(select_local_max(normalized_power[:,t],threshold))
        num_peaks = num_peaks + len(inds_above_thr)
        index.extend([num_peaks])
        if inds_above_thr.shape[0] == 0:
            continue

        freqs_peaks = freqs[inds_above_thr]        
        freqs_new.extend(freqs_peaks)
        times_new.extend(alltimes[t] * np.ones((len(freqs_peaks),1)))
        powss_new.extend(normalized_power[inds_above_thr,t])
        

    times_new = np.squeeze(times_new)
    freqs_new = np.squeeze(freqs_new)
    powss_new = np.squeeze(powss_new)
    
    return times_new,freqs_new,powss_new,index

def remove_doppler_from_peakmap(times_new,freqs_new,index,vec_n,vs,Nts):
    freqs_undop = []
    for t in range(Nts):
        if t == Nts:
            freqs_one_time = freqs_new[index[t]:]
        else:
            freqs_one_time = freqs_new[index[t]:index[t+1]]
            
        freqs_nodop = remove_doppler(freqs_one_time,vec_n,vs[:,t])
        freqs_undop.extend(freqs_nodop)
    
    return freqs_undop


def select_local_max(pows,threshold):

    peaks_index = list()
    for eqpow in range(1,len(pows)-1):
        if pows[eqpow]>threshold:
            if pows[eqpow]>pows[eqpow+1]:
                if pows[eqpow]>pows[eqpow-1]:
                    peaks_index.append(eqpow)
    return peaks_index

def remove_doppler_from_spectrogram_and_local_max_thresh(alltimes,freqs,normalized_power,vec_n,vs,threshold=3):
#     alltimes = times["H1"]
#     position = [signal_parameters["Alpha"], signal_parameters["Delta"]]
#     vec_n = astro2rect(position,1)
#     ts,vs = get_detector_velocities(data)
    Ntts = len(alltimes)
    freqs_new = []
    times_new = []
    powss_new = []
    for t in range(Ntts):
        inds_above_thr = np.array(select_local_max(normalized_power[:,t],threshold))
        if inds_above_thr.shape[0] == 0:
            continue
        freqs_peaks = freqs[inds_above_thr]
        freqs_dopp = remove_doppler(freqs_peaks,vec_n,vs[:,t])
        freqs_new.extend(freqs_dopp)
        times_new.extend(alltimes[t] * np.ones((len(freqs_dopp),1)))
        powss_new.extend(normalized_power[inds_above_thr,t])



    times_new = np.squeeze(times_new)
    freqs_new = np.squeeze(freqs_new)
    powss_new = np.squeeze(powss_new)
    
    return times_new,freqs_new,powss_new
    

def flatten_spectrogram(alltimes,freqs,normalized_power):
    Ntts = len(alltimes)
    freqs_flat = []
    times_flat = []
    powss_flat = []
    index = [0]
    num_peaks = 0
    for t in range(Ntts):       
        freqs_flat.extend(freqs)
        times_flat.extend(alltimes[t] * np.ones((len(freqs),1)))
        powss_flat.extend(normalized_power[:,t])
        

    times_flat = np.squeeze(times_flat)
    freqs_flat = np.squeeze(freqs_flat)
    powss_flat = np.squeeze(powss_flat)
    
    return times_flat,freqs_flat,powss_flat

def project_peaks(pm_freqs,pm_freqs_undopp):
    """
    Project a peak map onto the frequency axis.

    Parameters:
    - peaks: 2D array where peaks[1, :] contains the frequencies
    - plot_flag: 1 to plot the projected peak map; 0 to skip plotting

    Returns:
    - fbins: The frequency bins in the peak map
    - counts: The number of peaks per frequency bin in the projected peak map
    """
    # Get the unique frequency bins
    fbins = np.unique(pm_freqs)

    # Histogram the frequencies into the unique bins
    counts, fbins = np.histogram(pm_freqs_undopp, bins=fbins)


    return fbins[:-1], counts



import os
import github

for token in [
    "GH_TOKEN",
    "GH_TRAVIS_TOKEN",
    "GH_DRONE_TOKEN",
    "ORGWIDE_GH_TRAVIS_TOKEN",
]:
    try:
        gh = github.Github(os.environ[token])
        login = gh.get_user().login
    except Exception:
        loging = "NOT FOUND"

    print("%s: %s" % (token, login))



import numpy as np

def hfdf_hough_transients(peaks, hm_job):
    Day_inSeconds = 86400

    gridk = np.squeeze(hm_job['gridk'])
    minf0 = hm_job['minf']
    maxf0 = hm_job['maxf']
    df = hm_job['df']
    enh = hm_job['frenh']
    ref_perc_time = hm_job['ref_perc_time']
    braking_index = hm_job['n']
    poww = braking_index - 1

    if braking_index == 5 or braking_index == 3 or braking_index == 7:
        pass
    else:
        print('chirp, flipping spindowns to spinups')
        gridk = -gridk

    epoch = np.percentile(peaks[0, :], ref_perc_time*100)
    peaks[0, :] = peaks[0, :] - epoch

    if braking_index == 1:
        x = np.log(peaks[1, :])
        minx = np.min(x)
        maxx = np.max(x)
        dx = df * 1 / maxf0
        poww = 1 ### not physical, just to make codes work
    else:
        x = peaks[1, :] ** -poww
        minx = 1 / maxf0 ** poww
        maxx = 1 / minf0 ** poww

        if maxx < minx:
            minx, maxx = maxx, minx

        dx = (poww) * df * 1 / (maxf0) ** (braking_index)
        dx = abs(dx)

    ddx = dx / enh
    inix, finx = minx, maxx

    nbin_x = int(np.ceil((abs(finx - inix)) / ddx))

    ii = np.squeeze(np.argwhere(np.diff(peaks[0, :])))
    nTimeSteps = len(ii)

    ii0 = 0

    binh_df0 = np.zeros((len(gridk), nbin_x))

    for it in range(nTimeSteps):
        x0_a = ((x[ii0:ii[it]+1] - inix) / ddx)
        t = peaks[0, ii0] * Day_inSeconds
        tddx = t / ddx
        
        for id in range(len(gridk)):
            td = gridk[id] * tddx * poww
            inds = np.round(x0_a - td).astype(int)
            ind_of_inds = np.argwhere(inds >= 0)
            a = inds[ind_of_inds]
            log_inds = a <= nbin_x-1
            a = a[log_inds]
            binh_df0[id, a] = binh_df0[id,a] + 1

        ii0 = ii[it] + 1

    hm_job['epoch'] = epoch
    hm_job['dx'] = dx
    hm_job['gridx'] = np.arange(inix, finx, ddx)

    return binh_df0, hm_job

def make_hm_job_struct(minf, maxf, TFFT, dur, n, ref_perc_time, gridk):
    hm_job = {
        'minf': minf,               # minimum frequency to do the Hough on
        'maxf': maxf,               # maximum frequency to do the Hough on
        'df': 1 / TFFT,             # step in frequency
        'dur': dur,
        'patch': [0, 0],
        'n': n,
        'ref_perc_time': ref_perc_time,
        'frenh': 1,
        'gridk': gridk
    }

    return hm_job


def andrew_long_transient_grid_k(Tfft, f0range, f0dotrange, tobs, nb):
    f0min = min(f0range)
    f0max = max(f0range)
    f0dot_min = f0dotrange[0]
    f0dot_max = f0dotrange[1]
    log10fdotmin = np.log10(abs(f0dotrange[0]))
    log10fdotmax = np.log10(abs(f0dotrange[1]))

    randf = f0min + (f0max - f0min) * np.random.rand(10000)
    log10randfdot = log10fdotmin + (log10fdotmax - log10fdotmin) * np.random.rand(10000)

    each_k_step = []
    nk = []
    gridk = []

    for i in range(1):
        kmin = f0dot_min / f0max**nb
        kmax = f0dot_max / f0min**nb

        newk = kmax
        j = 1
        k = []

        while newk >= kmin:
            dk = newk * ((1 + 1 / (Tfft * f0max))**(-nb) - 1)
            each_k_step.append(dk)

            if j == 1:
                k.append(newk)
            else:
                k.append(newk + dk)

            newk = k[j - 1]
            j += 1

        nk.append(len(k))
        gridk.append(k)

    gridk = np.flip(gridk)
    each_k_step = np.abs(np.flip(each_k_step))

    return gridk, each_k_step

def cbc_shorten_gridk(gridk, mink, maxk, frac_around=0.15):
    factor = 1 + frac_around
    kmin = mink / factor
    kmax = maxk * factor
    inddd = np.argmin(np.abs(kmin - gridk))
    ind2 = np.argmin(np.abs(kmax - gridk))
    reduced_gridk = gridk[inddd:ind2 + 1]
    return reduced_gridk

def get_f0_from_x0(x0, n):
    f0 = x0**(-1 / (n - 1))
    return f0




import numpy as np

def constants():
    # fundamental constants
    c = 299792458  # speed of light
    h = 6.626e-34  # Planck's constant
    hbar = h / (2 * np.pi)
    ev = 1.602e-19  # electron volt
    Msun = 1.9885e30  # solar mass
    G = 6.67430e-11  # gravitational constant
    f_earth = 1 / 86400  # Earth's rotation frequency
    StellarDay = 23 * 3600 + 56 * 60 + 4.098903691  # length of a stellar day in seconds

    v0 = 0.000766667  # over c
    vesc = 0.00181459  # over c
    rhodm = 0.4e9 / 1e-6  # dark matter energy density (eV/m3)

    omega_earth_orb = 2 * np.pi / (365.25 * 86400)  # Earth's orbital angular frequency
    omega_earth_rot = 2 * np.pi * f_earth  # Earth's rotational angular frequency
    R_earth = 6371e3  # Earth's radius in meters
    Rorb = 149597871e3  # Earth's orbital radius in meters
    eps0 = 8.854187e-12  # vacuum permittivity
    fine_struct = 1 / 137  # fine structure constant

    units = {
        'ev_to_inv_s': ev / hbar,
        'ev_to_inv_m': ev / (hbar * c),
        'ev_to_kg': ev / c**2,
        'kg_to_ev': c**2 / ev,
        'charge_LH': ev / np.sqrt(4 * np.pi * fine_struct),
        'charge_G': ev / np.sqrt(fine_struct),
        'kpc_to_m': 3.086e19,
        'mpc_to_m': 3.086e22
    }

    consts = {
        'c': c,
        'h': h,
        'hbar': hbar,
        'hbar_inev': hbar / ev,
        'ev': ev,
        'Msun': Msun,
        'G': G,
        'eps0': eps0,
        'fine_struct': fine_struct,
        'f_earth': f_earth,
        'omega_earth_rot': omega_earth_rot,
        'omega_earth_orb': omega_earth_orb,
        'v_earth_rot': omega_earth_rot * R_earth,
        'v_earth_orb': omega_earth_orb * Rorb,
        'Rorb': Rorb,
        'Re': R_earth,
        'v0': v0,
        'vesc': vesc,
        'rhodm': rhodm,
        'StellarDay': StellarDay,
        'units': units
    }

    return consts

# Example usage:
# constants_data = constants()
# print(constants_data)

def calc_mc_with_k(k):
    # Assuming constants() is a function that returns the constants dictionary
    consts = constants()
    G = consts['G']
    c = consts['c']
    msun = consts['Msun']

    mc = k**(3/5) * (5 / (96 * np.pi**(8/3)))**(3/5) / (G / c**3)
    mc /= msun

    return mc




from shutil import rmtree
import tempfile
import conda.base.context
import conda.core.index
import conda.resolve
import conda_build.api
import conda_index.api
import networkx as nx
from compute_build_graph import construct_graph
import argparse
import atexit
import re
import os
from collections import OrderedDict
from pathlib import Path
import sys
import subprocess
import yaml

try:
    from ruamel_yaml import BaseLoader, load
except ImportError:
    from yaml import BaseLoader, load


EXAMPLE_RECIPE_FOLDERS = ["example", "example-v1"]
LOCAL_CHANNELS = os.environ.get("CONDA_BLD_PATH", "local").split(",")


def get_host_platform():
    from sys import platform

    if platform == "linux" or platform == "linux2":
        return "linux"
    elif platform == "darwin":
        return "osx"
    elif platform == "win32":
        return "win"


def get_config_name(arch):
    platform = get_host_platform()
    return os.environ.get("CONFIG", "{}{}".format(platform, arch))


def build_all(recipes_dir, arch):
    folders = [
        d
        for d in os.listdir(recipes_dir)
        if os.path.isdir(os.path.join(recipes_dir, d))
        and d not in EXAMPLE_RECIPE_FOLDERS
    ]
    if not folders:
        print("Found no recipes to build")
        return

    platform = get_host_platform()
    script_dir = os.path.dirname(os.path.realpath(__file__))
    variant_config_file = os.path.join(
        script_dir, "{}.yaml".format(get_config_name(arch))
    )

    has_meta_yaml = False
    has_recipe_yaml = False

    found_cuda = False
    found_centos7 = False
    for folder in folders:
        meta_yaml = os.path.join(recipes_dir, folder, "meta.yaml")
        if os.path.exists(meta_yaml):
            has_meta_yaml = True
            with open(meta_yaml, "r", encoding="utf-8") as f:
                text = "".join(f.readlines())
                if "cuda" in text:
                    found_cuda = True
                if "sysroot_linux-64" in text:
                    found_centos7 = True

        recipe_yaml = os.path.join(recipes_dir, folder, "recipe.yaml")
        if os.path.exists(recipe_yaml):
            has_recipe_yaml = True
            with open(recipe_yaml, "r", encoding="utf-8") as f:
                text = "".join(f.readlines())
                if "cuda" in text:
                    found_cuda = True
                if "sysroot_linux-64" in text:
                    found_centos7 = True

        cbc = os.path.join(recipes_dir, folder, "conda_build_config.yaml")
        if os.path.exists(cbc):
            with open(cbc, "r") as f:
                lines = f.readlines()
            pat = re.compile(
                r"^([^\#]*?)\s+\#\s\[.*(not\s(linux|unix)|(?<!not\s)(osx|win)).*\]\s*$"
            )
            # remove lines with selectors that don't apply to linux, i.e. if they contain
            # "not linux", "not unix", "osx" or "win"; this also removes trailing newlines
            lines = [pat.sub("", x) for x in lines]
            text = "\n".join(lines)
            if platform == "linux" and ("c_stdlib_version" in text):
                config = load(text, Loader=BaseLoader)
                if "c_stdlib_version" in config:
                    for version in config["c_stdlib_version"]:
                        version = tuple([int(x) for x in version.split(".")])
                        print(f"Found c_stdlib_version for linux: {version=}")
                        found_centos7 |= version == (2, 17)

    if has_meta_yaml and has_recipe_yaml:
        raise ValueError("Mixing meta.yaml and recipe.yaml recipes is not supported")
    if not has_meta_yaml and not has_recipe_yaml:
        raise ValueError("Neither a meta.yaml or a recipe.yaml recipes was found")

    if found_cuda:
        print("##vso[task.setvariable variable=NEED_CUDA;isOutput=true]1")
    if found_centos7:
        os.environ["DEFAULT_LINUX_VERSION"] = "cos7"
        print("Overriding DEFAULT_LINUX_VERSION to be cos7")

    deployment_version = (0, 0)
    sdk_version = (0, 0)
    channel_urls = None
    for folder in folders:
        cbc = os.path.join(recipes_dir, folder, "conda_build_config.yaml")
        if os.path.exists(cbc):
            with open(cbc, "r") as f:
                lines = f.readlines()
            pat = re.compile(
                r"^([^\#]*?)\s+\#\s\[.*(not\s(osx|unix)|(?<!not\s)(linux|win)).*\]\s*$"
            )
            # remove lines with selectors that don't apply to osx, i.e. if they contain
            # "not osx", "not unix", "linux" or "win"; this also removes trailing newlines
            lines = [pat.sub("", x) for x in lines]
            text = "\n".join(lines)
            if platform == "osx" and (
                "MACOSX_DEPLOYMENT_TARGET" in text
                or "MACOSX_SDK_VERSION" in text
                or "c_stdlib_version" in text
            ):
                config = load(text, Loader=BaseLoader)

                if "MACOSX_DEPLOYMENT_TARGET" in config:
                    for version in config["MACOSX_DEPLOYMENT_TARGET"]:
                        version = tuple([int(x) for x in version.split(".")])
                        deployment_version = max(deployment_version, version)
                if "c_stdlib_version" in config:
                    for version in config["c_stdlib_version"]:
                        version = tuple([int(x) for x in version.split(".")])
                        print(f"Found c_stdlib_version for osx: {version=}")
                        deployment_version = max(deployment_version, version)
                if "MACOSX_SDK_VERSION" in config:
                    for version in config["MACOSX_SDK_VERSION"]:
                        version = tuple([int(x) for x in version.split(".")])
                        sdk_version = max(sdk_version, deployment_version, version)

            if "channel_sources" not in text:
                new_channel_urls = [*LOCAL_CHANNELS, "conda-forge"]
            else:
                config = load(text, Loader=BaseLoader)
                new_channel_urls = [
                    *LOCAL_CHANNELS,
                    *config["channel_sources"][0].split(","),
                ]
            if channel_urls is None:
                channel_urls = new_channel_urls
            elif channel_urls != new_channel_urls:
                raise ValueError(
                    f"Detected different channel_sources in the recipes: {channel_urls} vs. {new_channel_urls}. Consider submitting them in separate PRs"
                )

    if channel_urls is None:
        channel_urls = [*LOCAL_CHANNELS, "conda-forge"]

    with open(variant_config_file, "r") as f:
        variant_text = "".join(f.readlines())

    if deployment_version != (0, 0):
        deployment_version = ".".join([str(x) for x in deployment_version])
        print("Overriding MACOSX_DEPLOYMENT_TARGET to be ", deployment_version)
        variant_text += "\nMACOSX_DEPLOYMENT_TARGET:\n"
        variant_text += f'- "{deployment_version}"\n'

    if sdk_version != (0, 0):
        sdk_version = ".".join([str(x) for x in sdk_version])
        print("Overriding MACOSX_SDK_VERSION to be ", sdk_version)
        variant_text += "\nMACOSX_SDK_VERSION:\n"
        variant_text += f'- "{sdk_version}"\n'

    with open(variant_config_file, "w") as f:
        f.write(variant_text)

    if platform == "osx" and (sdk_version != (0, 0) or deployment_version != (0, 0)):
        subprocess.run("run_conda_forge_build_setup", shell=True, check=True)

    if "conda-forge" not in channel_urls:
        raise ValueError("conda-forge needs to be part of channel_sources")

    if has_meta_yaml:
        print("Building {} with {}".format(",".join(folders), ",".join(channel_urls)))
        build_folders(recipes_dir, folders, arch, channel_urls)
    elif has_recipe_yaml:
        print(
            "Building {} with {} using rattler-build".format(
                ",".join(folders), ",".join(channel_urls)
            )
        )
        build_folders_rattler_build(recipes_dir, platform, arch, channel_urls)


def get_config(arch, channel_urls):
    exclusive_config_files = [
        os.path.join(conda.base.context.context.root_prefix, "conda_build_config.yaml")
    ]
    script_dir = os.path.dirname(os.path.realpath(__file__))
    # since variant builds override recipe/conda_build_config.yaml, see
    # https://github.com/conda/conda-build/blob/3.21.8/conda_build/variants.py#L175-L181
    # we need to make sure not to use variant_configs here, otherwise
    # staged-recipes PRs cannot override anything using the recipe-cbc.
    exclusive_config_file = os.path.join(
        script_dir, "{}.yaml".format(get_config_name(arch))
    )
    if os.path.exists(exclusive_config_file):
        exclusive_config_files.append(exclusive_config_file)

    config = conda_build.api.Config(
        arch=arch,
        exclusive_config_files=exclusive_config_files,
        channel_urls=channel_urls,
        error_overlinking=True,
    )
    return config


def build_folders(recipes_dir, folders, arch, channel_urls):
    index_path = os.path.join(sys.exec_prefix, "conda-bld")
    os.makedirs(index_path, exist_ok=True)
    conda_index.api.update_index(index_path)
    index = conda.core.index.get_index(channel_urls=channel_urls)
    conda_resolve = conda.resolve.Resolve(index)

    config = get_config(arch, channel_urls)
    platform = get_host_platform()

    worker = {
        "platform": platform,
        "arch": arch,
        "label": "{}-{}".format(platform, arch),
    }

    G = construct_graph(
        recipes_dir,
        worker=worker,
        run="build",
        conda_resolve=conda_resolve,
        folders=folders,
        config=config,
        finalize=False,
    )
    order = list(nx.topological_sort(G))
    order.reverse()

    print(
        "Computed that there are {} distributions to build from {} recipes".format(
            len(order), len(folders)
        )
    )
    if not order:
        print("Nothing to do")
        return
    print("Resolved dependencies, will be built in the following order:")
    print("    " + "\n    ".join(order))

    d = OrderedDict()
    for node in order:
        d[G.nodes[node]["meta"].meta_path] = 1

    for recipe in d.keys():
        conda_build.api.build([recipe], config=get_config(arch, channel_urls))


def build_folders_rattler_build(
    recipes_dir: str, platform, arch, channel_urls: list[str]
):
    config = get_config(arch, channel_urls)

    # Remove the example recipes to ensure that they are not also built.
    for example_recipe in EXAMPLE_RECIPE_FOLDERS:
        rmtree(os.path.join(recipes_dir, example_recipe), ignore_errors=True)

    # Determine the locations for the variant config files.
    specs = OrderedDict()
    for f in config.exclusive_config_files:
        specs[f] = conda_build.variants.parse_config_file(
            os.path.abspath(os.path.expanduser(os.path.expandvars(f))), config
        )

    # Combine all the variant config files together
    combined_spec = conda_build.variants.combine_specs(specs, log_output=config.verbose)
    variant_config = yaml.dump(combined_spec)

    # Define the arguments for rattler-build
    args = [
        "rattler-build",
        "build",
        "--recipe-dir",
        recipes_dir,
        "--target-platform",
        f"{platform}-{arch}",
    ]

    # Construct a temporary file where we write the combined variant config. We can then pass that
    # to rattler-build.
    with tempfile.NamedTemporaryFile(delete=False) as fp:
        fp.write(variant_config.encode("utf-8"))
        atexit.register(os.unlink, fp.name)

    # Execute rattler-build.
    subprocess.run(args + ["--variant-config", fp.name], check=True)


def check_recipes_in_correct_dir(root_dir, correct_dir):
    for path in Path(root_dir).glob("*"):
        path = Path(path)
        if path.is_dir() and path.name.lower() in (
            ".pixi",
            "build_artifacts",
            "miniforge3",
        ):
            # ignore pkg_cache in build_artifacts
            continue
        for recipe_path in path.rglob("*.yaml"):
            if recipe_path.name not in ("meta.yaml", "recipe.yaml"):
                continue
            recipe_path = recipe_path.absolute().relative_to(root_dir)
            if (
                recipe_path.parts[0] != correct_dir
                and recipe_path.parts[0] != "broken-recipes"
            ) or len(recipe_path.parts) != 3:
                raise RuntimeError(
                    f"recipe {recipe_path} in wrong directory; "
                    f"must be under {correct_dir}/<name>/"
                )


def read_mambabuild(recipes_dir):
    """
    Only use mambabuild if all the recipes require so via
    'conda_build_tool: mambabuild' in 'recipes/<recipe>/conda-forge.yml'
    """
    folders = os.listdir(recipes_dir)
    conda_build_tools = []
    for folder in folders:
        if folder in EXAMPLE_RECIPE_FOLDERS:
            continue
        cf = os.path.join(recipes_dir, folder, "conda-forge.yml")
        if os.path.exists(cf):
            with open(cf, "r") as f:
                cfy = yaml.safe_load(f.read())
            conda_build_tools.append(cfy.get("conda_build_tool", "conda-build"))
        else:
            conda_build_tools.append("conda-build")
    if conda_build_tools:
        return all([tool == "mambabuild" for tool in conda_build_tools])
    return False


def use_mambabuild():
    from boa.cli.mambabuild import prepare

    prepare()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--arch",
        default="64",
        help="target architecture (second component of a subdir; e.g. 64, arm64, ppc64le)",
    )
    args = parser.parse_args()
    root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    check_recipes_in_correct_dir(root_dir, "recipes")
    use_mamba = read_mambabuild(os.path.join(root_dir, "recipes"))
    if use_mamba:
        use_mambabuild()
        subprocess.run("conda clean --all --yes", shell=True, check=True)
    build_all(os.path.join(root_dir, "recipes"), args.arch)



#!/bin/sh
"""exec" "python3" "$0" "$@" #"""  # fmt: off # fmt: on
# The line above this comment is a bash / sh / zsh guard
# to stop people from running it with the wrong interpreter

import glob
import os
import platform
import subprocess
import sys
from argparse import ArgumentParser
from subprocess import check_output


def verify_system():
    branch_name = check_output(
        ["git", "rev-parse", "--abbrev-ref", "HEAD"], text=True
    ).strip()
    if branch_name == "main":
        raise RuntimeError(
            "You should run build-locally from a new branch, not 'main'. "
            "Create a new one with:\n\n"
            "    git checkout -b your-chosen-branch-name\n"
        )


def setup_environment(ns):
    os.environ["CONFIG"] = ns.config
    os.environ["UPLOAD_PACKAGES"] = "False"
    os.environ["IS_PR_BUILD"] = "True"
    if ns.debug:
        os.environ["BUILD_WITH_CONDA_DEBUG"] = "1"
        if ns.output_id:
            os.environ["BUILD_OUTPUT_ID"] = ns.output_id
    if "MINIFORGE_HOME" not in os.environ:
        os.environ["MINIFORGE_HOME"] = os.path.join(
            os.path.dirname(__file__), "miniforge3"
        )
    if "OSX_SDK_DIR" not in os.environ:
        os.environ["OSX_SDK_DIR"] = os.path.join(os.path.dirname(__file__), "SDKs")

    # The default cache location might not be writable using docker on macOS.
    if ns.config.startswith("linux") and platform.system() == "Darwin":
        os.environ["CONDA_FORGE_DOCKER_RUN_ARGS"] = (
            "-e RATTLER_CACHE_DIR=/tmp/rattler_cache"
        )


def run_docker_build(ns):
    script = ".scripts/run_docker_build.sh"
    subprocess.check_call([script])


def run_osx_build(ns):
    script = ".scripts/run_osx_build.sh"
    subprocess.check_call([script])


def run_win_build(ns):
    script = ".scripts/run_win_build.bat"
    subprocess.check_call(["cmd", "/D", "/Q", "/C", f"CALL {script}"])


def verify_config(ns):
    choices_filter = ns.filter or "*"
    valid_configs = {
        os.path.basename(f)[:-5]
        for f in glob.glob(f".ci_support/{choices_filter}.yaml")
    }
    if choices_filter != "*":
        print(f"filtering for '{choices_filter}.yaml' configs")
    print(f"valid configs are {valid_configs}")
    if ns.config in valid_configs:
        print("Using " + ns.config + " configuration")
        return
    elif len(valid_configs) == 1:
        ns.config = valid_configs.pop()
        print("Found " + ns.config + " configuration")
    elif ns.config is None:
        print("config not selected, please choose from the following:\n")
        selections = list(enumerate(sorted(valid_configs), 1))
        for i, c in selections:
            print(f"{i}. {c}")
        try:
            s = input("\n> ")
        except KeyboardInterrupt:
            print("\nno option selected, bye!", file=sys.stderr)
            sys.exit(1)
        idx = int(s) - 1
        ns.config = selections[idx][1]
        print(f"selected {ns.config}")
    else:
        raise ValueError("config " + ns.config + " is not valid")
    if ns.config.startswith("osx") and platform.system() == "Darwin":
        if "OSX_SDK_DIR" not in os.environ:
            raise RuntimeError(
                "Need OSX_SDK_DIR env variable set. Run 'export OSX_SDK_DIR=/opt'"
                "to download the SDK automatically to '/opt/MacOSX<ver>.sdk'"
            )


def main(args=None):
    p = ArgumentParser("build-locally")
    p.add_argument("config", default=None, nargs="?")
    p.add_argument(
        "--filter",
        default=None,
        help="Glob string to filter which build choices are presented in interactive mode.",
    )
    p.add_argument(
        "--debug",
        action="store_true",
        help="Setup debug environment using `conda debug`",
    )
    p.add_argument("--output-id", help="If running debug, specify the output to setup.")

    ns = p.parse_args(args=args)
    verify_system()
    verify_config(ns)
    setup_environment(ns)

    if ns.config.startswith("linux") or (
        ns.config.startswith("osx") and platform.system() == "Linux"
    ):
        run_docker_build(ns)
    elif ns.config.startswith("osx"):
        run_osx_build(ns)
    elif ns.config.startswith("win"):
        run_win_build(ns)


if __name__ == "__main__":
    main()
